use defmt::*;
use embassy_futures::select::*;
use embassy_rp::gpio;
use embassy_rp::pac;
use embassy_rp::interrupt::UART0_IRQ;
use embassy_time::{Duration, Timer};
use embassy_embedded_hal::SetConfig;
use embassy_rp::peripherals::{self, UART0};
use embassy_rp::uart::{self, UartRx};
use embassy_rp::pio::{Config, Pio, ShiftConfig, ShiftDirection, FifoJoin};
use embassy_rp::relocate::RelocatedProgram;
use embassy_rp::Peripheral;
use {defmt_rtt as _, panic_probe as _};
use fixed::traits::ToFixed;
use fixed_macro::types::U56F8;
use heapless::Vec;
use crc::{Crc, CRC_16_MODBUS};

use crate::dont_abort::DontAbort;
use crate::dont_abort::DontAbortMode::*;

pub struct RS485 {
    pio: peripherals::PIO0,
    rx_pin: peripherals::PIN_17,
    tx_pin: peripherals::PIN_16,
    tx_en_pin: peripherals::PIN_15,
    dma_channel: peripherals::DMA_CH0,
    tx_dma_channel: peripherals::DMA_CH2,
    rx: UartRx<'static, UART0, uart::Async>
}

fn pin_io<P: gpio::Pin>(pin: &P) -> pac::io::Gpio {
    use gpio::Bank;

    let block = match pin.bank() {
        Bank::Bank0 => pac::IO_BANK0,
        Bank::Qspi => pac::IO_QSPI,
    };
    block.gpio(pin.pin() as _)
}

#[allow(dead_code)]
async fn debug_print_pio_addr(sm: pac::pio::StateMachine) {
    let mut prev = 42u8;
    loop {
        let addr = unsafe { sm.addr().read().addr() };
        if prev != addr {
            prev = addr;
            info!("SM addr: {}", addr);
        }
        Timer::after(Duration::from_millis(200)).await;
    }
}

#[repr(u8)]
pub enum ModbusErrorCode {
    IllegalFunction = 1,
    IllegalDataAddress = 2,
    IllegalDataValue = 3,
    ServerDeviceFailure = 4,
    Acknowledge = 5,
    ServerDeviceBusy = 6,
    MemoryParityError = 7,
    GatewayPathUnavailable = 0xa,
    GatewayTargetDeviceFailedToRespond = 0xb,
}

fn read_modbus_holding_register(addr: u16) -> Result<u16, ModbusErrorCode> {
    if addr == 1 {
        return Ok(42)
    }
    Err(ModbusErrorCode::IllegalDataAddress)
}

fn read_modbus_input_register(addr: u16) -> Result<u16, ModbusErrorCode> {
    if addr == 2 {
        return Ok(42)
    }
    Err(ModbusErrorCode::IllegalDataAddress)
}

fn write_modbus_register(addr: u16, value: u16) -> Result<(), ModbusErrorCode> {
    Err(ModbusErrorCode::IllegalDataAddress)
}

fn modbus_reply_error(rxbuf: &Vec<u8, 32>, txbuf: &mut Vec<u8, 32>, code: ModbusErrorCode) {
    txbuf.clear();
    txbuf.push(rxbuf[0]).unwrap();
    txbuf.push(rxbuf[1] | 0x80).unwrap();
    txbuf.push(code as u8).unwrap();
}

fn handle_modbus_frame2(rxbuf: &Vec<u8, 32>, txbuf: &mut Vec<u8, 32>) -> Result<(), ModbusErrorCode> {
    use ModbusErrorCode::*;

    info!("Modbus frame: {:?}", rxbuf.as_slice());

    match rxbuf[1] {
        0x03 => {
            // read holding registers
            if rxbuf.len() != 8 {
                // we shouldn't get here
                return Err(ServerDeviceFailure);
            }
            let start = ((rxbuf[2] as u16) << 8) | rxbuf[3] as u16;
            let quantity = ((rxbuf[4] as u16) << 8) | rxbuf[5] as u16;
            if quantity as usize > (txbuf.capacity() - 5) / 2 || quantity >= 128 {
                return Err(IllegalDataValue); // is that right?
            }
            txbuf.push(rxbuf[0]).or(Err(ServerDeviceFailure))?;
            txbuf.push(rxbuf[1]).or(Err(ServerDeviceFailure))?;
            txbuf.push((quantity*2) as u8).or(Err(ServerDeviceFailure))?;
            for i in 0..quantity {
                let value = read_modbus_holding_register(start + i)?;
                txbuf.push((value >> 8) as u8).or(Err(ServerDeviceFailure))?;
                txbuf.push((value & 0xff) as u8).or(Err(ServerDeviceFailure))?;
            }
            Ok(())
        },
        0x04 => {
            // read input registers
            if rxbuf.len() != 8 {
                // we shouldn't get here
                return Err(ServerDeviceFailure);
            }
            let start = ((rxbuf[2] as u16) << 8) | rxbuf[3] as u16;
            let quantity = ((rxbuf[4] as u16) << 8) | rxbuf[5] as u16;
            if quantity as usize > (txbuf.capacity() - 5) / 2 || quantity >= 128 {
                return Err(IllegalDataValue); // is that right?
            }
            txbuf.push(rxbuf[0]).or(Err(ServerDeviceFailure))?;
            txbuf.push(rxbuf[1]).or(Err(ServerDeviceFailure))?;
            txbuf.push((quantity*2) as u8).or(Err(ServerDeviceFailure))?;
            for i in 0..quantity {
                let value = read_modbus_input_register(start + i)?;
                txbuf.push((value >> 8) as u8).or(Err(ServerDeviceFailure))?;
                txbuf.push((value & 0xff) as u8).or(Err(ServerDeviceFailure))?;
            }
            Ok(())
        },
        0x06 => {
            // write register
            Err(IllegalDataAddress)
        },
        0x10 => {
            // write multiple registers
            Err(IllegalDataAddress)
        },
        _ => {
            Err(IllegalFunction)
        },
    }
}

fn handle_modbus_frame(rxbuf: &Vec<u8, 32>, txbuf: &mut Vec<u8, 32>) {
    match handle_modbus_frame2(rxbuf, txbuf) {
        Ok(()) => {
            if txbuf.capacity() - txbuf.len() < 2 {
                // We don't have enough space for the CRC so reply with error instead.
                modbus_reply_error(rxbuf, txbuf, ModbusErrorCode::ServerDeviceFailure);
            }
        },
        Err(code) => {
            modbus_reply_error(rxbuf, txbuf, code);
        }
    }

    const CRC: Crc<u16> = Crc::<u16>::new(&CRC_16_MODBUS);
    let x = CRC.checksum(txbuf.as_slice());
    txbuf.push((x & 0xff) as u8).unwrap();
    txbuf.push((x >> 8) as u8).unwrap();
}

enum ModbusFrameLength {
    NeedMoreData(u16),
    Length(u16),
    Unknown,
}

//FIXME This won't work if this is a response frame!
fn get_modbus_frame_length(rxbuf: &[u8]) -> ModbusFrameLength {
    use ModbusFrameLength::*;

    if rxbuf.len() < 3 {
        return NeedMoreData(3);
    }

    match rxbuf[1] {
        0x01..=0x06 => Length(8),
        0x0f | 0x10 => if rxbuf.len() == 7 { Length(9 + rxbuf[6] as u16) } else { NeedMoreData(7) },
        _ => Unknown,
    }
}

impl RS485 {
    pub fn new(
            uart: UART0, rx_pin: peripherals::PIN_17, tx_pin: peripherals::PIN_16, tx_en_pin: peripherals::PIN_15,
            rx_irq: UART0_IRQ, rx_dma_channel: peripherals::DMA_CH1, uart_config: uart::Config,
            pio: peripherals::PIO0, dma_channel: peripherals::DMA_CH0, tx_dma_channel: peripherals::DMA_CH2
        ) -> RS485 {
        // SAFETY: The auto-baud will only read from this pin and we will set it back to UART mode after initialising the PIO.
        let rx_pin_for_autobaud = unsafe { rx_pin.clone_unchecked() };

        let uart_rx = UartRx::new(
            uart,
            rx_pin,
            rx_irq,
            rx_dma_channel,
            uart_config,
        );
        RS485 { pio, rx_pin: rx_pin_for_autobaud, tx_pin, tx_en_pin, dma_channel, tx_dma_channel, rx: uart_rx }
    }

    pub async fn run_task(mut self: Self) {
        let Pio {
            mut common,
            sm0: mut sm,
            mut sm1,
            mut irq0,
            mut irq1,
            ..
        } = Pio::new(self.pio);

        sm.set_enable(false);

        let prg_set_osr = pio_proc::pio_asm!(
            "pull",
            "hang:",
            "jmp hang",
        );

        let prg_autobaud = pio_proc::pio_asm!(
            // Send zero to RX FIFO to signal start of next measurement.
            ".wrap_target",
            "timeout:",
            "in null, 32",

            // Wait for pin to be high for a longer time (UART is idle), copy waiting time to X
            // and decrement until done, start again when pin changes.
            "wait_idle:",
            "wait 1 pin 0",
            "mov x, osr",
            "continue_wait_idle:",
            "jmp pin still_one",
            "jmp wait_idle",  // we got a zero, start again
            "still_one:",
            "jmp x--, continue_wait_idle",

            // We have a long 1, i.e. UART is idle. Wait for start bit.
            "have_long_high:",
            "irq set 1",          // notify MCU in case they want to know about this (end of Modbus frame)
            "wait 0 pin 0",
            "nop [4]",
            "jmp pin wait_idle",  // abort if zero was just a glitch

            // We want to measure 9 bits (start bit plus eight (or nine) data bits). However, we need
            // different code for high and low so it's easier to measure 4x2 bits. Weuse Y as the loop counter.
            "set y, 3",

            // Start counting. We count down from the value in OSR (which is our timeout).
            "measure_next:",
            "mov x, osr",
            "measure_low:",
            "jmp pin changed_to_one",
            //"nop",
            "jmp x-- measure_low",
            "jmp timeout",
            "changed_to_one:",
            "in x, 32"

            // Start counting, again, but for a high pulse.
            "mov x, osr",
            "measure_high:",
            "jmp pin still_high2",
            "jmp changed_to_low",
            "still_high2:",
            "jmp x-- measure_high",
            "jmp timeout",
            "changed_to_low:",
            "in x, 32",
            "jmp y-- measure_next",

            // We are done.
            "irq set 0",
            ".wrap",
        );

        // This is based on the UART TX example in the datasheet but we also control the TX_EN pin
        // and we have 9 data bits because we have parity.
        /*
        let prg_tx = pio_proc::pio_asm!(
            ".side_set 2 opt",
            // An 8e1 UART transmit program.
            // OUT pin 0 and side-set pin 1 are both mapped to UART TX pin.
            "pull side 1",           // disable transmitter, get number of bytes
            "mov y, osr side 3",     // copy to y reg, enable transmitter
            "byteloop:",
            "pull side 3 [6]",       // Assert stop bit, or stall with line in idle state (6 instead of 7 because mov/jmp take one clk)
            "set x, 8 side 2 [7]",   // Preload bit counter, assert start bit for 8 clocks
            "bitloop:",              // This loop will run 9 times (8e1 UART, i.e. 8 data bits and one clock bit)
            "out pins, 1",           // Shift 1 bit from OSR to the first OUT pin
            "jmp x-- bitloop [6]",   // Each loop iteration is 8 cycles.
            "jmp y-- byteloop side 3",
        );
        */
        // Let's use 4 cycles for each baud time because we need two bits of side-set.
        let prg_tx = pio_proc::pio_asm!(
            ".origin 24",
            ".side_set 2 opt",
            // An 8e1 UART transmit program.
            // OUT pin 0 and side-set pin 1 are both mapped to UART TX pin.
            "pull side 2",           // disable transmitter, get number of bytes. Side: TX_EN=0, TX=1
            "mov y, osr side 3",     // copy to y reg, enable transmitter
            "byteloop:",
            "pull side 3 [2]",       // Assert stop bit, or stall with line in idle state (6 instead of 7 because mov/jmp take one clk)
            "set x, 8 side 1 [3]",   // Preload bit counter, assert start bit for 8 clocks. Side: TX_EN=1, TX=0
            "bitloop:",              // This loop will run 9 times (8e1 UART, i.e. 8 data bits and one clock bit)
            "out pins, 1",           // Shift 1 bit from OSR to the first OUT pin
            "jmp x-- bitloop [2]",   // Each loop iteration is 8 cycles.
            "jmp y-- byteloop side 3",
        );

        let relocated = RelocatedProgram::new(&prg_set_osr.program);
        let mut cfg = Config::default();
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[]);
        const SM_FREQ: u32 = 25_000_000;
        cfg.clock_divider = (U56F8!(125_000_000) / U56F8!(25_000_000 /* SM_FREQ */)).to_fixed();
        cfg.shift_in = ShiftConfig {
            auto_fill: true,
            threshold: 32,
            direction: ShiftDirection::Left,
        };
        cfg.shift_out = ShiftConfig {
            auto_fill: false,
            threshold: 32,
            direction: ShiftDirection::Right,
        };
        let rx_pin_pio = common.make_pio_pin(unsafe { self.rx_pin.clone_unchecked() });
        cfg.set_in_pins(&[&rx_pin_pio]);
        cfg.set_jmp_pin(&rx_pin_pio);

        sm.set_config(&cfg);
        sm.set_enable(true);

        // Timeout: Modbus wants 3.5 char times between frames but we should detect the end of a frame
        // after 1.5 char times. This is 0.86 ms at 19200 baud (8e1, i.e. 11 baud times per char).
        // We need two clocks for each loop.
        let timeout_start_value = (SM_FREQ as f32 * 0.86e-3 / 2.) as u32;
        sm.tx().push(timeout_start_value);
        info!("timeout_start_value: {} = 0x{:08x}", timeout_start_value, timeout_start_value);

        // switch to the real program and join FIFOs
        unsafe { common.free_instr(loaded_program.used_memory); };
        sm.set_enable(false);
        let relocated = RelocatedProgram::new(&prg_autobaud.program);
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[]);
        cfg.fifo_join = FifoJoin::RxOnly;
        sm.set_config(&cfg);
        sm.set_enable(true);

        // set rx pin function back to UART
        // (PIO can always read and we don't want to write but embassy_rp cannot know that so it "helpfully" configured the pin for PIO function.)
        unsafe {
            pin_io(&self.rx_pin).ctrl().write(|w| w.set_funcsel(embassy_rp::pac::io::vals::Gpio17ctrlFuncsel::UART0_RX.0));
        }

        info!("Program size for auto-baud: {}, for tx: {}", prg_autobaud.program.code.len(), prg_tx.program.code.len());


        let relocated = RelocatedProgram::new(&prg_tx.program);
        let mut cfg = Config::default();
        let tx_pin_pio = common.make_pio_pin(self.tx_pin);
        let tx_en_pin_pio = common.make_pio_pin(self.tx_en_pin);
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[&tx_en_pin_pio, &tx_pin_pio]);
        const BAUD_RATE: u32 = 19200;
        cfg.clock_divider = (U56F8!(125_000_000) / U56F8!(76800 /* BAUD_RATE*4 */)).to_fixed();
        cfg.shift_out = ShiftConfig {
            auto_fill: false,
            threshold: 32,
            direction: ShiftDirection::Right,
        };
        cfg.set_out_pins(&[&tx_pin_pio]);
        cfg.fifo_join = FifoJoin::TxOnly;
        sm1.set_config(&cfg);
        sm1.set_pins(gpio::Level::Low, &[&tx_en_pin_pio]);
        sm1.set_pins(gpio::Level::High, &[&tx_pin_pio]);
        sm1.set_pin_dirs(embassy_rp::pio::Direction::Out, &[&tx_en_pin_pio, &tx_pin_pio]);
        sm1.set_enable(true);

        //FIXME Can we split Modbus parts from UART/RS485?
        let mut rxbuf = Vec::<u8, 32>::new();
        const CRC: Crc<u16> = Crc::<u16>::new(&CRC_16_MODBUS);
        let mut rxcrc = CRC.digest();
        let mut rx_expected_bytes = ModbusFrameLength::NeedMoreData(3);
        let mut rx_received_bytes = 0u16;

        const TX_BUF_LENGTH: usize = 32;
        let mut txbuf = Vec::<u8, TX_BUF_LENGTH>::new();
        let mut tx_data = [0; TX_BUF_LENGTH+1];

        let mut dma_in_ref = self.dma_channel.into_ref();
        let mut dma_tx_ref = self.tx_dma_channel.into_ref();
        let mut din = [42u32; 9];
        let mut bit_index = 0;
        let mut rx_buf_one = [0; 1];
        let mut rx_future = DontAbort::new(self.rx.read(&mut rx_buf_one), PanicIfReused);
        // We transmit with length zero to have a dummy no-op future with the right type. This seems to work ok.
        let mut tx_future = DontAbort::new(sm1.tx().dma_push(dma_tx_ref.reborrow(), &tx_data[0..0]), HangIfReused);
        loop {
            let x = select4(
                irq0.wait(),
                sm.rx().dma_pull(dma_in_ref.reborrow(), &mut din),
                //debug_print_pio_addr(embassy_rp::pac::PIO0.sm(0)),
                tx_future.continue_wait(),
                select(
                    rx_future.continue_wait(),
                    //debug_print_pio_addr(embassy_rp::pac::PIO0.sm(1)),
                    core::future::pending::<()>(),
                ),
            ).await;
            match x {
                Either4::First(_) => {
                    let mut sum = 0;
                    let mut first = 0;
                    let mut ok = false;
                    let mut got_to_eight = false;
                    for i in 0..din.len() {
                        if din[i] == 0 {
                            bit_index = 0;
                            info!("SM in {}: {:08x} -> start", i, din[i]);
                        } else {
                            let mut delay = timeout_start_value - din[i];
                            if bit_index == 0 {
                                delay += 7;
                            } else if bit_index%2 == 1 {
                                //delay += 4;
                                delay += 3;
                            } else {
                                delay += 3;
                            }
                            let millis = (delay) as f32 / SM_FREQ as f32 * 1000. * 2.;
                            let baud = 1000. / millis;
                            info!("SM in {} ({}): {:08x} -> {} -> {} ms -> {}", i, bit_index, din[i], delay, millis, baud);
                            if bit_index == 0 {
                                sum = 0;
                                first = delay;
                                ok = true;
                            } else if delay < first-first/4 || delay > first+first/4 {
                                ok = false;
                            }
                            sum += delay;
                            bit_index += 1;
                            got_to_eight = bit_index == 8;
                        }
                    }

                    for _ in bit_index..8 {
                        if let Some(x) = sm.rx().try_pull() {
                            let mut delay = timeout_start_value - x;
                            if bit_index == 0 {
                                delay += 7;
                            } else if bit_index%2 == 1 {
                                //delay += 4;
                                delay += 3;
                            } else {
                                delay += 3;
                            }
                            let millis = (delay) as f32 / SM_FREQ as f32 * 1000. * 2.;
                            let baud = 1000. / millis;
                            info!("SM in ({}): {:08x} -> {} -> {} ms -> {}", bit_index, x, delay, millis, baud);
                            if bit_index == 0 {
                                sum = 0;
                                first = delay;
                                ok = true;
                            } else if delay < first-first/4 || delay > first+first/4 {
                                ok = false;
                            }
                            sum += delay;
                            bit_index += 1;
                            got_to_eight = bit_index == 8;
                        } else {
                            break
                        }
                    }

                    if got_to_eight && ok {
                        let millis = sum as f32 / SM_FREQ as f32 * 1000. * 2. / 8.;
                        let baud = 1000. / millis;
                        info!("Guessed baud rate: {}", baud);
                    }
                },
                Either4::Second(_) => {
                    if false {
                        for i in 0..din.len() {
                            if din[i] == 0 {
                                bit_index = 0
                            } else {
                                bit_index += 1
                            }
                            info!("SM in {}: {:08x}", i, din[i]);
                        }
                    }
                },
                Either4::Third(_) => {
                    //drop(tx_future);
                    //tx_future = DontAbort::new(sm1.tx().dma_push(dma_tx_ref.reborrow(), &tx_data), HangIfReused);
                },
                Either4::Fourth(Either::First(x)) => {
                    drop(rx_future);
                    match x {
                        Result::Ok(()) => {
                            info!("RX {:?}", rx_buf_one);
                            let rx_char = rx_buf_one[0];

                            rxcrc.update(&[rx_char]);
                            if !rxbuf.is_full() {
                                rxbuf.push(rx_char).unwrap_or_default();
                            }
                            rx_received_bytes += 1;

                            if let ModbusFrameLength::NeedMoreData(x) = rx_expected_bytes {
                                if x == rx_received_bytes {
                                    rx_expected_bytes = get_modbus_frame_length(rxbuf.as_slice());
                                    match rx_expected_bytes {
                                        ModbusFrameLength::Unknown => {
                                            //FIXME Wait for pause.
                                        },
                                        _ => {}
                                    }
                                }
                            }
                            if let ModbusFrameLength::Length(x) = rx_expected_bytes {
                                if x == rx_received_bytes {
                                    let calculated_crc = rxcrc.finalize();
                                    rxcrc = CRC.digest();  // re-init to placate borrow checker
                                    const CORRECT_CRC: u16 = 0;  // because we include the CRC bytes in our calculation
                                    if calculated_crc != CORRECT_CRC {
                                        info!("CRC: {:04x} (should be zero)", calculated_crc);
                                    }

                                    //FIXME In case of CRC mismatch, wait for gap/idle of >=1.5 chars.
                                    const OUR_ADDRESS: u8 = 1;
                                    if rxbuf[0] == OUR_ADDRESS && calculated_crc == CORRECT_CRC {
                                        txbuf.clear();
                                        handle_modbus_frame(&rxbuf, &mut txbuf);

                                        if !txbuf.is_empty() {
                                            drop(tx_future);

                                            tx_data[0] = (txbuf.len() - 1) as u32;
                                            for i in 0..txbuf.len() {
                                                let x = txbuf[i] & 0xff;
                                                let mut parity = 0;
                                                for j in 0..8 {
                                                    parity ^= x>>j;
                                                }
                                                tx_data[i + 1] = x as u32 | (((parity as u32) & 1) << 8);
                                            }

                                            info!("Modbus reply: {:?}", txbuf);

                                            tx_future = DontAbort::new(sm1.tx().dma_push(dma_tx_ref.reborrow(),
                                                &tx_data[0..(txbuf.len()+1)]), HangIfReused);
                                        }
                                    }

                                    rxbuf.clear();
                                    //rxcrc = CRC.digest();  // re-init already done above, don't repeat to avoid warning
                                    rx_expected_bytes = ModbusFrameLength::NeedMoreData(3);
                                    rx_received_bytes = 0;
                                }
                            }
                        },
                        Result::Err(e) => {
                            info!("RX error {:?}", e);

                            //FIXME wait for gap/idle of >=1.5 chars.
                            rxbuf.clear();
                            rxcrc = CRC.digest();
                            rx_expected_bytes = ModbusFrameLength::NeedMoreData(3);
                            rx_received_bytes = 0;
                        },
                    }
                    rx_future = DontAbort::new(self.rx.read(&mut rx_buf_one), PanicIfReused);
                },
                _ => {
                },
            }
        }
    }
}
